@@ -19,7 +19,6 @@ from datetime import datetime
 from astroplan.plots import plot_airmass, plot_altitude
 from ztfquery import fields, query  # type: ignore
 from shapely.geometry import Polygon  # type: ignore
-from typing import Union, Optional
 
 from planobs import gcn_parser, utils
 
@@ -28,6 +27,10 @@ ztf = ["ZTF", "ztf"]
 
 logger = logging.getLogger(__name__)
 
+SIGNALNESS_THRESHOLD = 0.5
+AREA_THRESHOLD = 10.0
+AREA_HARD_THRESHOLD = 40.0
+
 
 class PlanObservation:
     """
@@ -40,16 +43,17 @@ class PlanObservation:
     def __init__(
         self,
         name: str,
-        ra: Optional[float] = None,
-        dec: Optional[float] = None,
-        arrivaltime: str = None,
-        date: str = None,
-        max_airmass=2.0,
+        ra: float | None = None,
+        dec: float | None = None,
+        arrivaltime: str | None = None,
+        date: str | None = None,
+        max_airmass=1.9,
         observationlength: float = 300,
         bands: list = ["g", "r"],
         multiday: bool = False,
-        alertsource: str = None,
-        site: Union[str, Observer] = "Palomar",
+        alertsource: str | None = None,
+        site: str | Observer = "Palomar",
+        switch_filters: bool = False,
         verbose: bool = True,
         **kwargs,
     ) -> None:
@@ -62,10 +66,11 @@ class PlanObservation:
         self.observationlength = observationlength
         self.bands = bands
         self.multiday = multiday
-        self.ra: Optional[float] = None
-        self.dec: Optional[float] = None
-        self.ra_err: Optional[list] = None
-        self.dec_err: Optional[list] = None
+        self.switch_filters = switch_filters
+        self.ra: float | None = None
+        self.dec: float | None = None
+        self.ra_err: list | None = None
+        self.dec_err: list | None = None
         self.warning = None
         self.observable = True
         self.rejection_reason = None
@@ -82,7 +87,12 @@ class PlanObservation:
             # check if name is correct
             assert utils.is_icecube_name(self.name)
 
+            test = gcn_parser.parse_latest_gcn_notice()
+
             gcn_nr = gcn_parser.find_gcn_circular(neutrino_name=self.name)
+            notice = gcn_parser.parse_latest_gcn_notice()
+            self.signalness = notice["signalness"]
+            self.energy = notice["energy"]
 
             if gcn_nr:
 
@@ -110,16 +120,10 @@ class PlanObservation:
                     logger.info(
                         "The IceCube alert is from the same day as the latest GCN circular, there is probably no GCN circular available yet. Using latest GCN notice"
                     )
-                    (
-                        ra,
-                        dec,
-                        arrivaltime,
-                        revision,
-                    ) = gcn_parser.parse_latest_gcn_notice()
-                    self.ra = ra
-                    self.dec = dec
-                    self.arrivaltime = arrivaltime
-                    self.datasource = revision
+                    self.ra = notice["ra"]
+                    self.dec = notice["dec"]
+                    self.arrivaltime = notice["arrivaltime"]
+                    self.datasource = f"Notice {notice['revision']}\n"
 
                 else:
                     raise ValueError(
@@ -171,12 +175,12 @@ class PlanObservation:
 
         constraints = [
             ap.AltitudeConstraint(20 * u.deg, 90 * u.deg),
-            ap.AirmassConstraint(max_airmass),
+            ap.AirmassConstraint(self.max_airmass),
             ap.AtNightConstraint.twilight_astronomical(),
         ]
 
         # Obtain moon coordinates at Palomar for the full time window (default: 24 hours from running the script)
-        times = Time(self.start_obswindow + np.linspace(0, 24, 1000) * u.hour)
+        times = Time(self.start_obswindow + np.linspace(0, 24, 1440) * u.hour)
         moon_times = Time(self.start_obswindow + np.linspace(0, 24, 50) * u.hour)
         moon_coords = []
 
@@ -219,7 +223,7 @@ class PlanObservation:
                 if (
                     (t_mjd < self.twilight_morning.mjd - 0.03)
                     or (t_mjd > self.twilight_evening.mjd + 0.03)
-                ) and airmass[index] < 2.0:
+                ) and airmass[index] < self.max_airmass:
                     indices_included.append(index)
                     airmasses_included.append(airmass[index])
                     times_included.append(times[index])
@@ -227,7 +231,7 @@ class PlanObservation:
                 if (
                     (t_mjd > self.twilight_evening.mjd + 0.01)
                     and (t_mjd < self.twilight_morning.mjd - 0.01)
-                ) and airmass[index] < 2.0:
+                ) and airmass[index] < self.max_airmass:
                     indices_included.append(index)
                     airmasses_included.append(airmass[index])
                     times_included.append(times[index])
@@ -240,10 +244,22 @@ class PlanObservation:
             self.observable = False
             self.rejection_reason = "proximity to gal. plane"
 
-        self.g_band_recommended_time_start: Optional[astropy.time.core.Time] = None
-        self.g_band_recommended_time_end: Optional[astropy.time.core.Time] = None
-        self.r_band_recommended_time_start: Optional[astropy.time.core.Time] = None
-        self.r_band_recommended_time_end: Optional[astropy.time.core.Time] = None
+        if self.ra_err:
+
+            self.calculate_area()
+
+            if (
+                self.signalness < SIGNALNESS_THRESHOLD and self.area > AREA_THRESHOLD
+            ) or self.area >= AREA_HARD_THRESHOLD:
+                self.observable = False
+                self.rejection_reason = (
+                    f"(area: {self.area:.1f} sq. deg, sness={self.signalness:.2f})"
+                )
+
+        self.g_band_recommended_time_start: astropy.time.core.Time | None = None
+        self.g_band_recommended_time_end: astropy.time.core.Time | None = None
+        self.r_band_recommended_time_start: astropy.time.core.Time | None = None
+        self.r_band_recommended_time_end: astropy.time.core.Time | None = None
 
         if self.observable:
             min_airmass = np.min(airmasses_included)
@@ -253,48 +269,62 @@ class PlanObservation:
             distance_to_evening = min_airmass_time.mjd - self.twilight_evening.mjd
             distance_to_morning = self.twilight_morning.mjd - min_airmass_time.mjd
 
-            if distance_to_morning < distance_to_evening:
-                if "g" in self.bands:
-                    self.g_band_recommended_time_start = utils.round_time(
-                        min_airmass_time - self.observationlength * u.s - 0.5 * u.hour
-                    )
-                    self.g_band_recommended_time_end = (
-                        self.g_band_recommended_time_start
-                        + self.observationlength * u.s
-                    )
-                if "r" in self.bands:
-                    self.r_band_recommended_time_start = utils.round_time(
-                        min_airmass_time - self.observationlength * u.s
-                    )
-                    self.r_band_recommended_time_end = (
-                        self.r_band_recommended_time_start
-                        + self.observationlength * u.s
-                    )
+            # now we divide in two blocks of time if there are two bands required
+
+            if len(self.bands) == 2:
+                obs_window_length_min = (times_included[-1] - times_included[0]).to(
+                    u.min
+                )
+
+                # Create two blocks, separated by 2*8 minutes
+                divider = int(len(times_included) / 2)
+                obsblock_1 = times_included[0 : divider - 8]
+                obsblock_2 = times_included[divider + 8 :]
+
+                if distance_to_morning < distance_to_evening:
+                    g_band_obsblock = obsblock_1
+                    r_band_obsblock = obsblock_2
+                else:
+                    g_band_obsblock = obsblock_2
+                    r_band_obsblock = obsblock_1
+
+                self.g_band_recommended_time_start = utils.round_time(
+                    g_band_obsblock[0]
+                )
+                self.g_band_recommended_time_end = utils.round_time(g_band_obsblock[-1])
+                self.r_band_recommended_time_start = utils.round_time(
+                    r_band_obsblock[0]
+                )
+                self.r_band_recommended_time_end = utils.round_time(r_band_obsblock[-1])
 
             else:
-                if "g" in self.bands:
-                    self.g_band_recommended_time_start = utils.round_time(
-                        min_airmass_time + self.observationlength * u.s + 0.5 * u.hour
-                    )
-                    self.g_band_recommended_time_end = (
+                self.g_band_recommended_time_start = utils.round_time(times_included[0])
+                self.g_band_recommended_time_end = utils.round_time(times_included[-1])
+
+            if self.switch_filters:
+                if "g" in self.bands and "r" in self.bands:
+                    g_band_temp_start = self.r_band_recommended_time_start
+                    g_band_temp_end = self.r_band_recommended_time_end
+                    self.r_band_recommended_time_start = (
                         self.g_band_recommended_time_start
-                        + self.observationlength * u.s
-                    )
-                if "r" in self.bands:
-                    self.r_band_recommended_time_start = utils.round_time(
-                        min_airmass_time + self.observationlength * u.s
-                    )
-                    self.r_band_recommended_time_end = (
-                        self.r_band_recommended_time_start
-                        + self.observationlength * u.s
                     )
+                    self.r_band_recommended_time_end = self.g_band_recommended_time_end
+                    self.g_band_recommended_time_start = g_band_temp_start
+                    self.g_band_recommended_time_end = g_band_temp_end
+
         if self.alertsource in icecube:
             summarytext = f"Name = IceCube-{self.name[2:]}\n"
         else:
             summarytext = f"Name = {self.name}\n"
 
-        if self.ra_err is not None and self.dec_err is not None:
-            summarytext += f"RA = {self.coordinates.ra.deg} + {self.ra_err[0]} - {self.ra_err[1]*-1.0}\nDec = {self.coordinates.dec.deg} + {self.dec_err[0]} - {self.dec_err[1]*-1.0}\n"
+        if self.ra_err and self.dec_err:
+            if (
+                self.ra_err[0]
+                and self.ra_err[1]
+                and self.dec_err[0]
+                and self.dec_err[1]
+            ):
+                summarytext += f"RA = {self.coordinates.ra.deg} + {self.ra_err[0]} - {self.ra_err[1]*-1.0}\nDec = {self.coordinates.dec.deg} + {self.dec_err[0]} - {self.dec_err[1]*-1.0}\n"
         else:
             summarytext += f"RADEC = {self.coordinates.ra.deg:.8f} {self.coordinates.dec.deg:.8f}\n"
 
@@ -313,7 +343,7 @@ class PlanObservation:
         if self.site.name == "Palomar":
 
             if self.observable and not self.multiday:
-                summarytext += "Recommended observation times:\n"
+                summarytext += "Recommended observation windows:\n"
                 if "g" in self.bands:
                     gbandtext = f"g-band: {utils.short_time(self.g_band_recommended_time_start)} - {utils.short_time(self.g_band_recommended_time_end)} [UTC]"
                 if "r" in self.bands:
@@ -348,6 +378,19 @@ class PlanObservation:
 
         self.summarytext = summarytext
 
+    def calculate_area(self):
+        """Calculate the on-sky area from sky location and location errors"""
+
+        ra1 = self.ra + self.ra_err[0]
+        ra2 = self.ra + self.ra_err[1]
+        dec1 = self.dec + self.dec_err[0]
+        dec2 = self.dec + self.dec_err[1]
+        self.area = np.abs(
+            (180 / np.pi) ** 2
+            * (np.radians(ra2) - np.radians(ra1))
+            * (np.sin(np.radians(dec2)) - np.sin(np.radians(dec1)))
+        )
+
     def plot_target(self):
         """
         Plot the observation window, including moon, altitude
@@ -523,10 +566,15 @@ class PlanObservation:
             plt.legend()
 
         if self.observable is False:
+
+            if "area" in self.rejection_reason:
+                reason_header = "ABOVE QUALITY THRESHOLD\n"
+            else:
+                reason_header = "NOT OBSERVABLE\ndue to "
             plt.text(
                 0.5,
                 0.5,
-                f"NOT OBSERVABLE\ndue to {self.rejection_reason}",
+                reason_header + f"{self.rejection_reason}",
                 size=20,
                 rotation=30.0,
                 ha="center",
@@ -577,8 +625,7 @@ class PlanObservation:
         fieldids_ref = []
 
         if load_refs_from_archive:
-            mt = utils.get_references(fieldids)
-            print(mt)
+            mt: pd.DataFrame | None = utils.get_references(fieldids)
 
         else:
             zq = query.ZTFQuery()
@@ -594,11 +641,12 @@ class PlanObservation:
             zq.load_metadata(kind="ref", sql_query=querystring)
             mt = zq.metatable
 
-        if len(mt) > 0:
-            for f in mt.field.unique():
-                d = {k: k in mt["filtercode"].values for k in ["zg", "zr", "zi"]}
-                if d["zg"] == True and d["zr"] == True:
-                    fieldids_ref.append(int(f))
+        if mt is not None:
+            if len(mt) > 0:
+                for f in mt.field.unique():
+                    d = {k: k in mt["filtercode"].values for k in ["zg", "zr", "zi"]}
+                    if d["zg"] == True and d["zr"] == True:
+                        fieldids_ref.append(int(f))
 
         logger.info(f"Fields that contain target: {fieldids}")
         logger.info(f"Of these have a reference: {fieldids_ref}")
@@ -617,9 +665,15 @@ class PlanObservation:
         ccds = fields._CCD_COORDS
 
         coverage = {}
+        distance = {}
 
         for f in self.fieldids_ref:
             centroid = fields.get_field_centroid(f)
+            centroid_coords = SkyCoord(
+                centroid[0][0] * u.deg, centroid[0][1] * u.deg, frame="icrs"
+            )
+            dist_to_target = self.coordinates.separation(centroid_coords).deg
+            distance.update({f: dist_to_target})
 
             fig, ax = plt.subplots(dpi=300)
 
@@ -670,15 +724,16 @@ class PlanObservation:
             plt.close()
 
         self.coverage = coverage
+        self.distance = distance
 
         if len(self.coverage) > 0:
-
-            max_coverage_field = max(coverage, key=coverage.get)
+            max_coverage_field = max(self.coverage, key=self.coverage.get)
 
             self.recommended_field = max_coverage_field
 
         else:
-            self.recommended_field = None
+            # no errors -> no coverage -> let's use the more central field
+            self.recommended_field = min(self.distance, key=self.distance.get)
 
     def plot_finding_chart(self):
         """ """
